/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: LoaderManagerImpl.java 507 2006-06-02 07:43:59Z ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.classloader;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectStreamClass;
import java.lang.reflect.Field;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.List;
import java.util.Map;

import org.objectweb.petals.PetalsException;

/**
 * @version $Rev: 507 $ $Date: 2006-06-02 09:43:59 +0200 (ven, 02 jun 2006) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>, Adrien
 *         LOUIS
 */
public class LoaderManagerImpl implements LoaderManager {

    /**
     * Default constructor.
     */
    public LoaderManagerImpl() {
        super();
    }

    // ----------------------------------------------------------------------
    // ClassLoadingService interface implementation
    // ----------------------------------------------------------------------

    /**
     * Clears the caches maintained by the SunVM object stream implementation.
     * This method uses reflection and setAccessable to obtain access to the Sun
     * cache. The cache is locked with a synchronize monitor and cleared. This
     * method completely clears the class loader cache which will impact
     * preformance of object serialization.
     * 
     * @param clazz
     *            the name of the class containing the cache field
     * @param fieldName
     *            the name of the cache field
     */
    public static void clearSunSoftCache(Class clazz, String fieldName) {
        Map cache = null;
        try {
            Field field = clazz.getDeclaredField(fieldName);
            field.setAccessible(true);
            cache = (Map) field.get(null);
        } catch (Throwable ignored) {
            // there is nothing a user could do about this anyway
        }

        if (cache != null) {
            synchronized (cache) {
                cache.clear();
            }
        }
    }

    /**
     * Cleans well known class loader leaks in VMs and libraries. There is a lot
     * of bad code out there and this method will clear up the know problems.
     * This method should only be called when the class loader will no longer be
     * used. It this method is called two often it can have a serious impact on
     * preformance.
     * 
     * @param classLoader
     *            the class loader to destroy
     */
    public static void destroy() {
        clearSunSoftCache(ObjectInputStream.class, "subclassAudits");
        clearSunSoftCache(ObjectOutputStream.class, "subclassAudits");
        clearSunSoftCache(ObjectStreamClass.class, "localDescs");
        clearSunSoftCache(ObjectStreamClass.class, "reflectors");
    }

    /**
     * @see org.objectweb.petals.classloader.LoaderManager#containsClassLoader(java.lang.String)
     */
    public boolean containsClassLoader(String classLoaderId) {

        Loaders loaders = Loaders.getInstance();

        return loaders.contains(classLoaderId);
    }

    /**
     * Create a classLoader for a JBI Component. The parent classLoader will be
     * the SharedLibrariesClassLoader.
     * 
     * @see org.objectweb.petals.classloader.LoaderManager#createClassLoader(java.net.URL[],
     *      boolean)
     */
    public URLClassLoader createClassLoader(URL[] baseUrls,
            List<String> classpathLocations, boolean useParentFirst)
        throws PetalsException {

        // create the class loader
        PetalsClassLoader cl = null;

        Loaders loaders = Loaders.getInstance();

        try {
            cl = new PetalsClassLoader(baseUrls, classpathLocations, loaders
                    .getSharedLibrariesClassLoader());

            cl.setParentFirst(useParentFirst);
        } catch (IOException e) {
            throw new PetalsException("Error while creating a class loader", e);
        }
        return cl;
    }

    /**
     * @see org.objectweb.petals.classloader.LoaderManager#createClassLoader(java.lang.String,
     *      java.net.URL[], boolean)
     */
    public URLClassLoader createComponentClassLoader(String classLoaderId,
            URL[] baseUrls, List<String> classpathUrls, boolean useParentFirst,
            List<String> sharedLibrariesNameList) throws PetalsException {

        // create the class loader
        PetalsClassLoader cl = null;

        Loaders loaders = Loaders.getInstance();

        try {
            cl = new ComponentClassLoader(baseUrls, classpathUrls, loaders
                    .getSharedLibrariesClassLoader(), sharedLibrariesNameList);

            cl.setParentFirst(useParentFirst);
        } catch (IOException e) {
            throw new PetalsException("Error while creating a class loader", e);
        }

        // save for further retrieve (overwrite existing class loader)
        loaders.put(classLoaderId, cl);

        return cl;
    }

    /**
     * Create a classloader for a sharedLibrary. This classLoader will be added
     * to the SharedLibrariesClassLoader.
     * 
     * @see org.objectweb.petals.classloader.LoaderManager#createClassLoader(java.lang.String,
     *      java.net.URL[], boolean)
     */
    public URLClassLoader createSharedLibraryClassLoader(String classLoaderId,
            URL[] baseUrls, List<String> classpathUrls, boolean useParentFirst)
        throws PetalsException {

        // create the class loader
        SharedLibraryClassLoader cl = null;

        try {
            cl = new SharedLibraryClassLoader(baseUrls, classpathUrls);
            cl.setParentFirst(useParentFirst);
        } catch (IOException e) {
            throw new PetalsException("Error while creating a class loader", e);
        }

        // save for further retrieve (overwrite existing class loader)
        Loaders loaders = Loaders.getInstance();

        loaders.put(classLoaderId, cl);
        loaders.getSharedLibrariesClassLoader().addSharedLibraryClassLoader(
                classLoaderId, cl);

        return cl;
    }

    /**
     * Remove classLoader from the list. If the classLoader is a ClassLoader for
     * sharedLibraries, remove it from the SharedLibrariesClassLoader
     * 
     * @see org.objectweb.petals.classloader.LoaderManager#deleteClassLoader(java.lang.String)
     */
    public void deleteClassLoader(String classLoaderId) {

        Loaders loaders = Loaders.getInstance();

        if (!containsClassLoader(classLoaderId)) {
            return;
        }

        // delete class loader for further retrieve
        loaders.remove(classLoaderId);
        loaders.getSharedLibrariesClassLoader().removeSharedLibraryClassLoader(
                classLoaderId);
        /*
         * try { if (classLoader!=null)classLoader.close(); } catch (Throwable
         * e) { // TODO log e.printStackTrace(); }
         */
        destroy();
    }

}
